//*******************************************************
//       Copyright (c) 2003 Gennady A. Petrov
//*******************************************************
// Freeware v1.0
//
//
unit FCResStrings;

{$I FCCompilers.inc}

{$IFDEF LINUX}
  Not Worked in LINUX, only for MSWindows
{$ENDIF}

interface

uses
  Classes;

type
  TResStrItem = class(TCollectionItem)
  private
    FStringValue: String;
    FStringName: String;
  protected
    function GetDisplayName: string; override;
  public
    constructor Create(Collection: TCollection); override;
    procedure Assign(Source: TPersistent); override;
  published
    property StringName: String read FStringName write FStringName;
    property StringValue: String read FStringValue write FStringValue;
  end;

  TResStrItems = class(TCollection)
  private
    FControl: TComponent;
  protected
    procedure SetItem(Index: Integer; Value: TResStrItem);
    function GetItem(Index: Integer): TResStrItem;
    function Insert(Index: Integer): TResStrItem;
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(Control: TComponent);

    procedure Assign(Source: TPersistent); override;

    function Add: TResStrItem;
    property Items[Index: Integer]: TResStrItem read GetItem write SetItem; default;

    procedure Clear;
  end;

  TFCResStrings = class(TComponent)
  private
    FRSItems: TResStrItems;

    procedure SetResStrings(Value: TResStrItems);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property ResStrings: TResStrItems read FRSItems write SetResStrings;
  end;


  { TStrResList class }
  TStrResList = class;

  PStrResItemRec = ^TStrResItemRec;
  TStrResItemRec = packed record
    StrName: string;
    Base: PResStringRec;
    DefaultIdentifier: Integer;
  end;

  PStrResItemArray = ^TStrResItemArray;
  TStrResItemArray = array[0..MaxListSize] of TStrResItemRec;

  TStrResListSortCompare = function(List: TStrResList; Index1, Index2: Integer): Integer;

  TStrResList = class(TList)
  private
    FList: PStrResItemArray;
    FCount: Integer;
    FCapacity: Integer;
    procedure ExchangeItems(Index1, Index2: Integer);
    procedure Grow; override;
    procedure QuickSort(L, R: Integer; SCompare: TStrResListSortCompare);
    procedure InsertItem(Index: Integer; SR: TStrResItemRec);
  protected
    function Get(Index: Integer): TStrResItemRec;
    function GetCapacity: Integer;
    procedure Put(Index: Integer; SR: TStrResItemRec);
    procedure SetCapacity(NewCapacity: Integer);
  public
    property List: PStrResItemArray read FList;
    property Count: Integer read FCount write FCount;
    destructor Destroy; override;
    function Add(SR: TStrResItemRec): Integer;
    procedure Clear; override;
    procedure Delete(Index: Integer);
    procedure Exchange(Index1, Index2: Integer);
    function Find(SR: String; var Index: Integer): Boolean;
    procedure Insert(Index: Integer; SR: TStrResItemRec);
    procedure Sort; virtual;
    procedure CustomSort(Compare: TStrResListSortCompare); virtual;
  end;

  TFCResStrHolder = class(TObject)
  private
    FSRLItems: TStrResList;
    procedure Update(Value: TResStrItems);
  public
    property SRLItems: TStrResList read FSRLItems write FSRLItems;

    constructor Create(AOwner: TComponent);
    destructor Destroy; override;
    procedure SetStringValue(const SName: String; pBaseResStr: PResStringRec);

    procedure LoadDefault;
  end;

var
  FCResStrHolder: TFCResStrHolder;

implementation

uses
  {$IFDEF DELPHI6UP}
  RTLConsts,
  {$ENDIF}
  Windows,
  SysUtils;

destructor TStrResList.Destroy;
begin
  inherited Destroy;
  if FCount <> 0 then Finalize(FList^[0], FCount);
  FCount := 0;
  SetCapacity(0);
end;

function TStrResList.Add(SR: TStrResItemRec): Integer;
begin
  Result := FCount;
  InsertItem(Result, SR);
end;

procedure TStrResList.Clear;
begin
  if FCount <> 0 then
  begin
    Finalize(FList^[0], FCount);
    FCount := 0;
    SetCapacity(0);
  end;
end;

procedure TStrResList.Delete(Index: Integer);
begin
  {$IFDEF DELPHI6UP}
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  {$ENDIF}
  Finalize(FList^[Index]);
  Dec(FCount);
  if Index < FCount then
    System.Move(FList^[Index + 1], FList^[Index],
      (FCount - Index) * SizeOf(TStrResItemRec));
end;

procedure TStrResList.Exchange(Index1, Index2: Integer);
begin
  {$IFDEF DELPHI6UP}
  if (Index1 < 0) or (Index1 >= FCount) then Error(@SListIndexError, Index1);
  if (Index2 < 0) or (Index2 >= FCount) then Error(@SListIndexError, Index2);
  {$ENDIF}

  ExchangeItems(Index1, Index2);
end;

procedure TStrResList.ExchangeItems(Index1, Index2: Integer);
var
  Temp: Integer;
  Item1, Item2: PStrResItemRec;
begin
  Item1 := @FList^[Index1];
  Item2 := @FList^[Index2];

  Temp := Integer(Item1^.StrName);
  Integer(Item1^.StrName) := Integer(Item2^.StrName);
  Integer(Item2^.StrName) := Temp;

  Temp := Integer(Item1^.Base);
  Integer(Item1^.Base) := Integer(Item2^.Base);
  Integer(Item2^.Base) := Temp;

  Temp := Item1^.DefaultIdentifier;
  Item1^.DefaultIdentifier := Item2^.DefaultIdentifier;
  Item2^.DefaultIdentifier := Temp;
end;

function TStrResList.Find(SR: String; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := FCount - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := AnsiCompareText(FList^[I].StrName, SR);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        L := I;
      end;
    end;
  end;
  Index := L;
end;

function TStrResList.Get(Index: Integer): TStrResItemRec;
begin
  {$IFDEF DELPHI6UP}
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  {$ENDIF}

  Result := FList^[Index];
end;

function TStrResList.GetCapacity: Integer;
begin
  Result := FCapacity;
end;

procedure TStrResList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then Delta := FCapacity div 4 else
    if FCapacity > 8 then Delta := 16 else
      Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

procedure TStrResList.Insert(Index: Integer; SR: TStrResItemRec);
begin
  {$IFDEF DELPHI6UP}
  if (Index < 0) or (Index > FCount) then Error(@SListIndexError, Index);
  {$ENDIF}
  InsertItem(Index, SR);
end;

procedure TStrResList.InsertItem(Index: Integer; SR: TStrResItemRec);
begin
  if FCount = FCapacity then Grow;
  if Index < FCount then
    System.Move(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(TStrResItemRec));
  with FList^[Index] do
  begin
    Pointer(StrName) := nil;
    Base := SR.Base;
    DefaultIdentifier := SR.DefaultIdentifier;
    StrName := SR.StrName;
  end;

  Inc(FCount);
end;

procedure TStrResList.Put(Index: Integer; SR: TStrResItemRec);
begin
  {$IFDEF DELPHI6UP}
  if (Index < 0) or (Index >= FCount) then Error(@SListIndexError, Index);
  {$ENDIF}

  FList^[Index] := SR;
end;

procedure TStrResList.QuickSort(L, R: Integer; SCompare: TStrResListSortCompare);
var
  I, J, P: Integer;
begin
  repeat
    I := L;
    J := R;
    P := (L + R) shr 1;
    repeat
      while SCompare(Self, I, P) < 0 do Inc(I);
      while SCompare(Self, J, P) > 0 do Dec(J);
      if I <= J then
      begin
        ExchangeItems(I, J);
        if P = I then
          P := J
        else if P = J then
          P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J, SCompare);
    L := I;
  until I >= R;
end;

procedure TStrResList.SetCapacity(NewCapacity: Integer);
begin
  ReallocMem(FList, NewCapacity * SizeOf(TStrResItemRec));
  FCapacity := NewCapacity;
end;

function StrResListAnsiCompare(List: TStrResList; Index1, Index2: Integer): Integer;
begin
  Result := AnsiCompareText(List.FList^[Index1].StrName,
                            List.FList^[Index2].StrName);
end;

procedure TStrResList.Sort;
begin
  CustomSort(StrResListAnsiCompare);
end;

procedure TStrResList.CustomSort(Compare: TStrResListSortCompare);
begin
  if (FCount > 1) then
  begin
    QuickSort(0, FCount - 1, Compare);
  end;
end;

constructor TFCResStrHolder.Create(AOwner: TComponent);
begin
  FSRLItems := TStrResList.Create;
end;

destructor TFCResStrHolder.Destroy;
begin
  FSRLItems.Free;

  inherited;
end;

procedure TFCResStrHolder.Update(Value: TResStrItems);
var
  l,
  i: Integer;

  SaveFlag: Integer;
begin
  FSRLItems.Sort;

  for i := 0 to Value.Count-1 do
  begin
    if FSRLItems.Find(Value.Items[i].StringName, l) then
    begin
      with FSRLItems.FList^[l], Value.Items[i] do
      begin
        if VirtualProtect(Base, SizeOf(TResStringRec), PAGE_READWRITE, @SaveFlag) then
        try
          Base^.Identifier := Integer(StringValue);
        finally
          VirtualProtect(Base, SizeOf(TResStringRec), SaveFlag, @SaveFlag);
        end;
      end;
    end
  end;
end;

procedure TFCResStrHolder.SetStringValue(const SName: String; pBaseResStr: PResStringRec);
var
  SR: TStrResItemRec;
begin
  SR.StrName := SName;
  SR.Base := pBaseResStr;

  SR.DefaultIdentifier := SR.Base^.Identifier;
  FSRLItems.Add(SR);
end;

procedure TFCResStrHolder.LoadDefault;
var
  i: Integer;
  SaveFlag: Integer;
begin
  with FSRLItems do
    for i := FCount-1 downto 0 do
    begin
      with FList[i] do
      begin
        if VirtualProtect(Base, SizeOf(TResStringRec), PAGE_READWRITE, @SaveFlag) then
        try
          Base^.Identifier := DefaultIdentifier;
        finally
          VirtualProtect(Base, SizeOf(TResStringRec), SaveFlag, @SaveFlag);
        end;
      end;
    end;
end;

constructor TResStrItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);

  FStringName := '';
end;

procedure TResStrItem.Assign(Source: TPersistent);
begin
  if Source is TResStrItem then
  begin
    StringName:= TResStrItem(Source).StringName;
    StringValue:= TResStrItem(Source).StringValue;
  end
  else
    inherited Assign(Source);
end;

function TResStrItem.GetDisplayName: string;
begin
  Result := StringName;
  if Result = '' then Result := inherited GetDisplayName;
end;

constructor TResStrItems.Create(Control: TComponent);
begin
  inherited Create(TResStrItem);
  FControl := Control;
end;

function TResStrItems.Add: TResStrItem;
begin
  Result := TResStrItem.Create(Self);
end;

procedure TResStrItems.Assign(Source: TPersistent);
var
  I: Integer;
begin
  if Source is TResStrItems then
  begin

    BeginUpdate;
    try
      Clear;
      for I := 0 to TResStrItems(Source).Count - 1 do
        Add.Assign(TResStrItems(Source).Items[I]);
    finally
      EndUpdate;
    end;
    Exit;
  end;
  inherited Assign(Source);
end;

function TResStrItems.GetItem(Index: Integer): TResStrItem;
begin
  Result := TResStrItem(inherited GetItem(Index));
end;


function TResStrItems.Insert(Index: Integer): TResStrItem;
begin
  Result := Add;
  Result.Index := Index;
end;

procedure TResStrItems.SetItem(Index: Integer; Value: TResStrItem);
begin
  inherited SetItem(Index, Value);
end;

function TResStrItems.GetOwner: TPersistent;
begin
  Result := FControl;
end;

procedure TResStrItems.Update(Item: TCollectionItem);
begin
  inherited;
  if not (csDesigning in TFCResStrings(GetOwner).ComponentState) then
    FCResStrHolder.Update(Self);
end;


constructor TFCResStrings.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FRSItems:= TResStrItems.Create(Self);
end;

destructor TFCResStrings.Destroy;
begin
  FRSItems.Free;

  inherited Destroy;
end;

procedure TFCResStrings.SetResStrings(Value: TResStrItems);
begin
  FRSItems:= Value;
  FCResStrHolder.Update(Value);
end;

procedure TResStrItems.Clear;
var
  i: Integer;
begin
  BeginUpdate;
  for i:= Self.Count - 1 downto 0 do
  begin
    Delete(i);
  end;
  EndUpdate;
end;

initialization
  FCResStrHolder := TFCResStrHolder.Create(nil);

finalization
  FCResStrHolder.Free;

end.
